/* Cpu.java -	(c) Ilpo Jrvinen
 *
 * TiTo K2000 Harj.
 */



public class Cpu {

	public static final int	R0 = 0, R1 = 1, R2 = 2, R3 = 3, R4 = 4, R5 = 5, R6 = 6, R7 = 7;
	public static final int	FP = 7, SP = 6;		// alias
	public static final int PC = 8, IR = 9, SR = 10, TR = 11, BASE = 12, LIMIT = 13;

	public static final int MEMSIZE = 200;	
	public static final int REGSIZE = 14;
	
	private int		reg[] = new int [ REGSIZE ];
	private boolean	halted = true;
	private boolean tracing = false;
	
	private Instruction mem[] = new Instruction[ MEMSIZE ];
	
	public void tomem( int pos, Instruction i )
	{
		mem[ pos ] = i;
		if (tracing)
			System.out.println( "Mem: " + pos + " <= " + i );
	}
	
	public String memdump()
	{
		String s = "";
		for (int i = 0; i < MEMSIZE; i++)
			if (mem[i] != null)
				s += i + ":\t" + mem[i] + "\n";
//			else
//				s += 				// empty == "not used mem loc"
		return s;
	}
	
	public int	from( int r )
	{
		if (r < 0 || r > REGSIZE) {
			halt( "INVALID REGISTER, BUG IN CPU IMPLEMENTATION!" );
			return 0;
		}
		return reg[r];
	}
	
	public void exec( boolean trace )
	{
		tracing = trace;
		while (!halted) {
			executeOpcode( new Instruction( getMem( reg[ PC ]++ ) ) );
			if (tracing)
				System.out.println( "--- TRACE ---\n" + dump() );
		}
	}
	

	public void reset( int last, int base, int limit )
	{
		reg[ SP ] = last;
		reg[ FP ] = last - 2;
		reg[ BASE ] = base;
		reg[ LIMIT ] = limit;
		reg[ SR ] = 0;
		reg[ PC ] = 0;
		halted = false;
	}
	
	public void halt( String info )
	{
		halted = true;
		System.out.println( "--- HALT: " + info + " ---\n" + this );
		System.exit(1);
	}
	
	public String getFlags() 
	{
		String flags = "GEL OZUM IS P D";
		int j = 0;
		for (int i = 0; i < flags.length(); i++) {
			if (flags.charAt(i) == ' ')
				continue;

			// if flag is off, decapitalize the corresponding letter...
			if ((reg[ SR ] & (1 << j)) == 0) {
				flags = flags.substring(0,i) + Character.toLowerCase(flags.charAt(i)) + 
						flags.substring(i+1);
			}
			j++;
		}
		return flags;
	}
	
	public int getMem( int addr )
	{
		if (addr > reg[ LIMIT ]) {
			halt( "Address " + addr + " out of range" );
			return -1;
		}
		if (mem[ addr ] == null) {
//			halt( "Address " + addr + " doesn't contain valid value" );
			return 0;		// zero mem loc
		}
		return mem[ addr ].getValue();
	}
	
	public String dump()
	{
		return	"R0: " + reg[R0] + "\n" +
				"R1: " + reg[R1] + "\n" +
				"R2: " + reg[R2] + "\n" +
				"R3: " + reg[R3] + "\n" +
				"R4: " + reg[R4] + "\n" +
				"R5: " + reg[R5] + "\n" +
				"FP: " + reg[FP] + "\n" +
				"SP: " + reg[SP] + "\n" + 
				(!tracing ? "" :
				"PC: " + reg[PC] + "\n" +
				"SR: " + reg[SR] + "\n" +
				"TR: " + reg[TR] + "\n") +
				getFlags() + "\n";
	}
	
	public String toString()
	{
		String s = dump();
		s += memdump();
		return s;
	}

	/* 
	 * Method that gets long (64-bit) input, checks it for overflows.
	 * If result does NOT overflow it is converted back to 32-bits and
	 * returned to the caller, if overflow has occured then it execution
	 * is halted.
	 */
	public int ofCatch( long result )
	{
		if ((int)(result >> 32) != 0 && (int)(result >> 32) != -1) {
			reg[ SR ] |= 0x8;
			halt( "Overflow: " + result );
		}
		return (int)result;
	}
	
	public void executeOpcode( Instruction i )
	{
		switch ( i.getOp() ) {
			case Instruction.NOP:			// can't fail!
				return;
			case Instruction.LOAD:
				reg[ i.getRj() ] = i.getData( this );
				return;
			case Instruction.STORE:
				if (i.getM() > 1)
					halt( "Invalid addressing mode" );
				tomem( i.getData( this ), new Instruction( reg[ i.getRj() ] ) );
				return;
			case Instruction.ADD:
				reg[ i.getRj() ] = ofCatch( (long)reg[ i.getRj() ] + i.getData( this ) );
				break;
			case Instruction.SUB:
				reg[ i.getRj() ] = ofCatch( (long)reg[ i.getRj() ] - i.getData( this ) );
				break;
			case Instruction.MUL:
				reg[ i.getRj() ] = ofCatch( (long)reg[ i.getRj() ] * i.getData( this ) );
				break;
			case Instruction.DIV:
				if (i.getData( this ) == 0)
					halt( "Divide by zero" );
				reg[ i.getRj() ] = reg[ i.getRj() ] / i.getData( this );
				break;
			case Instruction.MOD:
				if (i.getData( this ) == 0)
					halt( "Divide by zero" );
				reg[ i.getRj() ] = reg[ i.getRj() ] % i.getData( this );
				break;
			case Instruction.JUMP:
				reg[ PC ] = i.getData( this );
				return;
			case Instruction.JZER:									// check cond.
				if (reg[ i.getRj() ] == 0)
					reg[ PC ] = i.getData( this );
				return;
			case Instruction.JGRE:									// check cond.
				if ((reg[ SR ] & 0x1) != 0)
					reg[ PC ] = i.getData( this );
				return;
			case Instruction.JEQU:									// check cond.
				if ((reg[ SR ] & 0x2) != 0)
					reg[ PC ] = i.getData( this );
				return;
			case Instruction.JLES:									// check cond.
				if ((reg[ SR ] & 0x4) != 0)
					reg[ PC ] = i.getData( this );
				return;
			case Instruction.SVC:
				if (i.getAddr() == Instruction.HALT && i.getM() == 0) {
					halt( "SVC" );
					return;								// EXEC FINISH!!!
				}
				halt( "Invalid SVC constant field: " + i.getAddr() );
				return;
			case Instruction.COMP:
				doFlags( reg[ i.getRj() ], i.getData( this ) );
				return;
			case Instruction.PUSH:
				reg[ i.getRj() ]++;
				tomem( reg[ i.getRj() ], new Instruction( i.getData( this ) ) );
				return;
			case Instruction.POP:
				if (i.getAddr() != 0 || i.getM() != 0) {
					halt( "Invalid POP" );
				}
				reg[ i.getRi() ] = i.getData( this );
				reg[ i.getRj() ]--;
				return;
			case Instruction.CALL:
				reg[ i.getRj() ]++;
				tomem( reg[ i.getRj() ], new Instruction( reg[ FP ] ) );
				reg[ i.getRj() ]++;
				tomem( reg[ i.getRj() ], new Instruction( reg[ PC ] ) );
				reg[ PC ] = i.getData( this );
				reg[ FP ] = reg[ SP ];
				return;
			case Instruction.EXIT:
				reg[ PC ] = i.getData( this );
				reg[ i.getRj() ]--;
				reg[ FP ] = i.getData( this );
				reg[ i.getRj() ] = reg[ i.getRj() ] - 1 - i.getData( this );
				return;
			default:
				halt( "Unknown opcode: " + i.toString() );
				return;
		}
		doFlags( reg[ i.getRj() ], 0 );
		return;
	}
	
	public void doFlags( int f, int s )
	{
//		System.out.println( "doing " + f + " vs " + s );
		if (f == s)
			reg[ SR ] = reg[ SR ] & ~0x7 | 0x2;
		else if (f < s)
			reg[ SR ] = reg[ SR ] & ~0x7 | 0x4;
		else
			reg[ SR ] = reg[ SR ] & ~0x7 | 0x1;
	}

}